#include "colors.h"
#include "matcom.h"
#include "pile.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum { NOT_VISITED = 0, IN_STACK = 1, DONE = 2 };

graph_mco mcgragh(void)
{
	graph_mco g;
	g.nbs = -1;
	g.nba = -1;
	g.func = malloc((g.fcap = 10) * sizeof(fix_str));
	assert(g.func);
	g.call = calloc((g.fcap = 10), sizeof(bool));
	assert(g.call);
	g.vec = malloc((g.vcap = 10) * sizeof(struct edge));
	assert(g.vec);
	g.vu = NULL;
	return g;
}

void mcinsert_func(graph_mco *g, fix_str i)
{
	++g->nbs;
	if (g->nbs == g->fcap) {
		g->func = realloc(g->func, (g->fcap += 128) * sizeof(fix_str));
		assert(g->func);
		g->call = realloc(g->call, g->fcap * sizeof(bool));
		assert(g->call);
	}
	strcpy(g->func[g->nbs], i);
}

void mcinsert_edge(graph_mco *g, int dep, int arr)
{
	struct edge e;
	++g->nba;
	if (g->nba == g->vcap) {
		g->vec = realloc(g->vec, (g->vcap += 128) * sizeof(struct edge));
		assert(g->vec);
	}
	e.i = dep;
	if (arr == -1) {
		e.j = g->nbs;
		g->call[g->nbs] = true;
	} else {
		e.j = arr;
		g->call[arr] = true;
	}
	g->vec[g->nba] = e;
}

void mcdelete_graph(graph_mco *g)
{
	if (g->func)
		free(g->func);
	if (g->vec)
		free(g->vec);
	if (g->vu)
		free(g->vu);
	if (g->call)
		free(g->call);

	g->func = NULL;
	g->vec = NULL;
	g->vu = NULL;
	g->call = NULL;
}

int mcis_in(graph_mco g, fix_str func)
{
	int i;
	if (g.nba == 0)
		return -1;
	for (i = 0; i <= g.nba; i++) {
		if (strcmp(g.func[i], func) == 0) {
			return i;
		}
	}
	return -1;
}

static void mcprint_succ(graph_mco g, int succ, int color)
{
	if (color)
		printf(YELLOW "  Succ:" RESET_COLOR CYAN " %s"RESET_COLOR"\n" , g.func[succ]);
	else
		printf("  Succ: %s\n" , g.func[succ]);
}

static void mcprint_edges(graph_mco g, int n, int color)
{
	int i, j = 0;
	if (g.vu[n])
		return;
	g.vu[n] = 1;
	if (color)
		printf(BLUE "NODE:" RESET_COLOR MAGENTA " %s\n" RESET_COLOR, g.func[n]);
	else
		printf("NODE: %s\n", g.func[n]);
	for (i = 0; i <= g.nba; ++i) {
		if (g.vec[i].i == n) {
			j++;
			mcprint_succ(g, g.vec[i].j, color);
		}
	}
	// HACK
	if (j == 0) {
		printf("\033[A\033[2K");
		rewind(stdout);
	}

	for (i = 0; i <= g.nba; ++i) {
		if (g.vec[i].i == n) {
			if (g.vec[g.vec[i].j].i != -1)
				mcprint_edges(g, g.vec[i].j, color);
		}
	}
}

void mcprint(graph_mco g, int n, int color)
{
	if (g.vu == NULL) {
		g.vu = calloc(g.nbs + 1, sizeof(int));
	}

	mcprint_edges(g, n, color);

	if (g.vu) {
		free(g.vu);
	}
}

static void write_node_dot(FILE *fp, graph_mco g)
{
	int i;
	for (i = 0; i <= g.nbs; i++) {
		fprintf(fp, "\t%s -> %s;\n", g.func[g.vec[i].i], g.func[g.vec[i].j]);
	}
}

static void main_graph(FILE *fp, graph_mco g, int n)
{
	int i;
	if (g.vu[n])
		return;
	g.vu[n] = 1;
	for (i = 0; i <= g.nba; ++i) {
		if (g.vec[i].i == n) {
			fprintf(fp, "\t%s -> %s;\n", g.func[n], g.func[g.vec[i].j]);
			main_graph(fp, g, g.vec[i].j);
		}
	}
}

void mcdot_graph(graph_mco *g, bool main)
{
	FILE *fp;
	int i;

	if (g->vu == NULL) {
		g->vu = calloc(g->nbs + 1, sizeof(int));
	} else {
		for (i = 0; i <= g->nbs; i++) {
			g->vu[i] = 0;
		}
	}

	fp = fopen("graph.dot", "a");
	// fp1 = fopen("mcCC.dot", "w");

	fprintf(fp, "digraph mcCCmain {\n\tlabelloc=\"t\";\n\tlabel=\"MAIN Compact Matrice\";\n");

	if (main) {
		fix_str m = "main";
		int n = mcis_in(*g, m);
		if (n != -1) {
			main_graph(fp, *g, n);
			fprintf(fp, "}\n");
		} else {
			fprintf(stderr, "main not found.\n");
			exit(1);
		}
		fprintf(fp, "digraph mcCC {\n\tlabelloc=\"t\";\n\tlabel=\"CC Compact Matrice\";\n");
		for (i = 0; i <= g->nbs; ++i) {
			if (!g->vu[i]) {
				main_graph(fp, *g, i);
			}
		}
		fprintf(fp, "}\n");
	} else {
		write_node_dot(fp, *g);
		fprintf(fp, "}\n");
	}


	fclose(fp);
}

// NOTE: https://www.baeldung.com/cs/detecting-cycles-in-directed-graph
static void print_cycle(graph_mco g, pile_t s, int v)
{
	pile_t s2 = init_pile();
	push(&s2, pop(&s));
	while (top(s2) != g.vec[v].j) {
		push(&s2, pop(&s));
	}
	while (!empty(s2)) {
		printf("%s -> ", g.func[pop(&s2)]);
	}
	printf("%s\n", g.func[g.vec[v].j]);
	free_pile(&s2);
}

static void processDFS(graph_mco g, int node, pile_t *s)
{
	int i;

	for (i = 0; i <= g.nba; ++i) {
		if (g.vec[i].i != node)
			continue;
		int v = g.vec[i].j;
		if (g.vu[v] == IN_STACK) {
			// puts("found");
			print_cycle(g, *s, i);
		} else if (g.vu[v] == NOT_VISITED) {
			push(s, v);
			g.vu[v] = IN_STACK;
			processDFS(g, v, s);
		}
	}
	g.vu[node] = DONE;
	pop(s);
}

void mcfind_cycles(graph_mco *g)
{
	int v, i;

	if (g->vu == NULL) {
		g->vu = calloc(g->nbs + 1, sizeof(int));
	} else {
		for (i = 0; i <= g->nbs; i++) {
			g->vu[i] = 0;
		}
	}

	for (v = 0; v <= g->nbs; ++v) {
		if (g->vu[v] == NOT_VISITED) {
			pile_t s = init_pile();
			push(&s, v);
			g->vu[v] = IN_STACK;
			processDFS(*g, v, &s);
			free_pile(&s);
		}
	}
}
