#include "colors.h"
#include "pile.h"
#include "vec.h"
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

enum { NOT_VISITED = 0, IN_STACK = 1, DONE = 2 };

graph_vec vnew_graph(void)
{
	graph_vec g;
	g.n = -1;
	g.n = -1;
	g.cap = 10;
	g.node = malloc(10 * sizeof(nodept));
	assert(g.node);
	return g;
}

void vinsert_node(graph_vec *g, fix_str str)
{
	++g->n;
	if (g->n == g->cap)
		g->node = realloc(g->node, (g->cap += 128) * sizeof(nodept));
	struct node *n = malloc(sizeof(struct node));
	assert(n);
	n->num = g->n;
	n->nbs = 0;
	n->vu = 0;
	n->cap = 5;
	n->call = false;
	n->succ = malloc(5 * sizeof(nodept));
	strcpy(n->func, str);
	g->node[g->n] = n;
}

void vdelete_graph(graph_vec *g)
{
	int i;
	for (i = 0; i <= g->n; ++i) {
		if (g->node[i]) {
			if (g->node[i]->succ) {
				free(g->node[i]->succ);
				g->node[i]->succ = NULL;
			}
			free(g->node[i]);
			g->node[i] = NULL;
		}
	}
	free(g->node);
}

int vis_in(graph_vec g, fix_str func)
{
	int i;
	if (g.n < 0) {
		return -1;
	}
	for (i = 0; i <= g.n; i++) {
		if (strcmp(g.node[i]->func, func) == 0) {
			return i;
		}
	}
	return -1;
}

static void vprint_succ(nodept node, int color)
{
	if (color)
		printf(YELLOW "  Succ:" RESET_COLOR CYAN "%s\n" RESET_COLOR, node->func);
	else
		printf("  Succ:%s\n", node->func);
}

static void vprint_nodes(nodept node, int color)
{
	int i;
	if (node->vu)
		return;
	node->vu = 1;
	if (color)
		printf(BLUE "NODE:" RESET_COLOR MAGENTA " %s\n" RESET_COLOR, node->func);
	else
		printf("NODE: %s\n", node->func);
	for (i = 0; i < node->nbs; ++i) {
		vprint_succ(node->succ[i], color);
	}
	for (i = 0; i < node->nbs; ++i) {
		if (node->succ[i]->nbs != 0)
			vprint_nodes(node->succ[i], color);
	}
}

void vprint(nodept node, graph_vec *g, int color)
{
	int i;
	for (i = 0; i <= g->n; i++) {
		g->node[i]->vu = 0;
	}
	vprint_nodes(node, color);
}

static void write_node_dot(FILE *fp, graph_vec g)
{
	int i, j;
	for (i = 0; i <= g.n; i++) {
		for (j = 0; j < g.node[i]->nbs; j++) {
			// printf("\t%s -> %s;\n",  node->func, node->succ[i]->func);
			fprintf(fp, "\t%s -> %s;\n", g.node[i]->func, g.node[i]->succ[j]->func);
		}
	}
}

static void main_graph(FILE *fp, nodept node)
{
	int i;
	if (node->vu == 1)
		return;
	node->vu = 1;
	// printf("NODE %d: %s\n", node->num, node->func);
	for (i = 0; i < node->nbs; ++i) {
		fprintf(fp, "\t%s -> %s;\n", node->func, node->succ[i]->func);
		main_graph(fp, node->succ[i]);
	}
}

void vdot_graph(graph_vec g, bool main)
{
	FILE *fp;
	int i;

	for (i = 0; i <= g.n; i++) {
		g.node[i]->vu = 0;
	}

	fp = fopen("graph.dot", "a");

	fprintf(fp, "digraph vCCmain {\n\tlabelloc=\"t\";\n\tlabel=\"MAIN Vector\";\n\trankdir=LR;\n");

	if (main) {
		fix_str m = "main";
		int n = vis_in(g, m);
		if (n != -1) {
			main_graph(fp, g.node[n]);
			fprintf(fp, "}\n");
		} else {
			fprintf(stderr, "main not found.\n");
			exit(1);
		}
		fprintf(fp, "digraph vCC {\n\tlabelloc=\"t\";\n\tlabel=\"CC Vector\";\n\trankdir=LR;\n");
		for (i = 0; i <= g.n; ++i) {
			if (!g.node[i]->call) {
				main_graph(fp, g.node[i]);
			}
		}

		fprintf(fp, "}\n");
	} else {
		write_node_dot(fp, g);
		fprintf(fp, "}\n");
	}

	fclose(fp);
}

// NOTE: https://www.baeldung.com/cs/detecting-cycles-in-directed-graph
static void print_cycle(graph_vec g, pile_t s, int v)
{
	pile_t s2 = init_pile();
	push(&s2, pop(&s));
	while (top(s2) != v) {
		push(&s2, pop(&s));
	}
	while (!empty(s2)) {
		printf("%s -> ", g.node[pop(&s2)]->func);
	}
	printf("%s\n", g.node[v]->func);
	free_pile(&s2);
}

static void processDFS(nodept n, pile_t *s, graph_vec g)
{
	int i;
	for (i = 0; i < n->nbs; ++i) {
		int v = n->succ[i]->num;
		if (n->succ[i]->vu == IN_STACK) {
			print_cycle(g, *s, v);
		} else if (n->succ[i]->vu == NOT_VISITED) {
			push(s, v);
			n->succ[i]->vu = IN_STACK;
			processDFS(n->succ[i], s, g);
		}
	}
	n->vu = DONE;
	pop(s);
}

void vfind_cycles(graph_vec *g)
{
	int v;
	for (v = 0; v <= g->n; v++) {
		g->node[v]->vu = 0;
	}
	for (v = 0; v <= g->n; ++v) {
		if (g->node[v]->vu == NOT_VISITED) {
			pile_t s = init_pile();
			push(&s, v);
			g->node[v]->vu = IN_STACK;
			processDFS(g->node[v], &s, *g);
			free_pile(&s);
		}
	}
}
